@@ -2,7 +2,7 @@
 import json
 import os
 import shutil
-import pickle as pkl
+import dill as pkl
 import tensorflow as tf
 import numpy as np
 
@@ -13,12 +13,13 @@ from tensorflow.keras.layers import Dense, Input
 from tensorflow.keras import regularizers
 from sklearn.base import BaseEstimator
 
-from .data.triplet_generators import generator_from_index, generator_from_knn_matrix
+
+from .data.generators import generator_from_neighbour_matrix, KerasSequence
 from .nn.network import triplet_network, base_network
 from .nn.callbacks import ModelCheckpoint
 from .nn.losses import triplet_loss, is_categorical, is_multiclass, is_hinge
 from .nn.losses import semi_supervised_loss, validate_sparse_labels
-from .data.knn import build_annoy_index
+from .data.neighbour_retrieval import AnnoyKnnMatrix
 
 
 class Ivis(BaseEstimator):
@@ -32,10 +33,18 @@ class Ivis(BaseEstimator):
     :param int embedding_dims: Number of dimensions in the embedding space
     :param int k: The number of neighbours to retrieve for each point.
         Must be less than one minus the number of rows in the dataset.
-    :param str distance: The loss function used to train the neural network.
-        One of "pn", "euclidean", "manhattan_pn", "manhattan", "chebyshev",
-        "chebyshev_pn", "softmax_ratio_pn", "softmax_ratio", "cosine",
-        "cosine_pn".
+    :param Union[str,Callable] distance: The loss function used to train
+        the neural network.
+
+        *   If string: a registered loss function name. Predefined losses are:
+            "pn", "euclidean", "manhattan_pn", "manhattan", "chebyshev",
+            "chebyshev_pn", "softmax_ratio_pn", "softmax_ratio", "cosine", "cosine_pn".
+        *   If Callable, must have two parameters, (y_true, y_pred).
+            y_pred denotes the batch of triplets, and y_true are any corresponding labels.
+            y_pred is expected to be of shape: (3, batch_size, embedding_dims).
+
+                * When loading model loaded with a custom loss, provide the loss to the
+                  constructor of the new Ivis instance before loading the saved model.
     :param int batch_size: The size of mini-batches used during gradient
         descent while training the neural network. Must be less than the
         num_rows in the dataset.
@@ -44,8 +53,6 @@ class Ivis(BaseEstimator):
         once.
     :param int n_epochs_without_progress: After n number of epochs without an
         improvement to the loss, terminate training early.
-    :param float margin: The distance that is enforced between points by the
-        triplet loss functions.
     :param int ntrees: The number of random projections trees built by Annoy to
         approximate KNN. The more trees the higher the memory usage, but the
         better the accuracy of results.
@@ -57,15 +64,17 @@ class Ivis(BaseEstimator):
     :param bool precompute: Whether to pre-compute the nearest neighbours.
         Pre-computing is a little faster, but requires more memory. If memory
         is limited, try setting this to False.
-    :param str model: str or keras.models.Model. The keras model to train using
-        triplet loss. If a model object is provided, an embedding layer of size
-        'embedding_dims' will be appended to the end of the network.
-        If a string, a pre-defined network by that name will be used. Possible
-        options are: 'szubert', 'hinton', 'maaten'. By default the 'szubert'
-        network will be created, which is a selu network composed of 3 dense
-        layers of 128 neurons each, followed by an embedding layer of size
-        'embedding_dims'.
-    :param str supervision_metric: str or function. The supervision metric to
+    :param Union[str,tf.keras.models.Model] model: The keras model to train using
+        triplet loss.
+
+        *   If a model object is provided, an embedding layer of size
+            'embedding_dims' will be appended to the end of the network.
+        *   If a string, a pre-defined network by that name will be used. Possible
+            options are: 'szubert', 'hinton', 'maaten'. By default the 'szubert'
+            network will be created, which is a selu network composed of 3 dense
+            layers of 128 neurons each, followed by an embedding layer of size
+            'embedding_dims'.
+    :param str supervision_metric: The supervision metric to
         optimize when training keras in supervised mode. Supports all of the
         classification or regression losses included with keras, so long as
         the labels are provided in the correct format. A list of keras' loss
@@ -78,15 +87,24 @@ class Ivis(BaseEstimator):
         saved on disk. If provided, the annoy index file will be used.
         Otherwise, a new index will be generated and saved to disk in the
         current directory as 'annoy.index'.
-    :param list[keras.callbacks.Callback] callbacks: List of keras Callbacks to
+    :param [keras.callbacks.Callback] callbacks: List of keras Callbacks to
         pass model during training, such as the TensorBoard callback. A set of
         ivis-specific callbacks are provided in the ivis.nn.callbacks module.
     :param bool build_index_on_disk: Whether to build the annoy index directly
         on disk. Building on disk should allow for bigger datasets to be indexed,
         but may cause issues.
-    :param np.array neighbour_matrix: A pre-computed KNN matrix can be provided.
-        The KNNs can be retrieved using any method, and will cause Ivis to skip
-        computing the Annoy KNN index.
+    :param Union[np.array,collections.abc.Sequence] neighbour_matrix:
+        Providing a neighbour matrix will cause Ivis to skip computing the Annoy KNN index
+        and instead use the provided neighbour_matrix.
+
+        - A pre-computed neighbour matrix can be provided as a numpy array. Indexing the array
+          should retrieve a list of neighbours for the data point associated with that index.
+
+        - Alternatively, dynamic computation of neighbours can be done by providing a
+          class than implements the collections.abc.Sequence class, specifically the
+          `__getitem__` and `__len__` methods.
+
+            - See the ivis.data.neighbour_retrieval.AnnoyKnnMatrix class for an example.
     :param int verbose: Controls the volume of logging output the model
         produces when training. When set to 0, silences outputs, when above 0
         will print outputs.
@@ -94,12 +112,12 @@ class Ivis(BaseEstimator):
     """
 
     def __init__(self, embedding_dims=2, k=150, distance='pn', batch_size=128,
-                 epochs=1000, n_epochs_without_progress=20,
-                 margin=1, ntrees=50, search_k=-1,
+                 epochs=1000, n_epochs_without_progress=20, ntrees=50,
+                 knn_distance_metric='angular', search_k=-1,
                  precompute=True, model='szubert',
                  supervision_metric='sparse_categorical_crossentropy',
                  supervision_weight=0.5, annoy_index_path=None,
-                 callbacks=None, build_index_on_disk=None,
+                 callbacks=None, build_index_on_disk=True,
                  neighbour_matrix=None, verbose=1):
 
         self.embedding_dims = embedding_dims
@@ -108,11 +126,11 @@ class Ivis(BaseEstimator):
         self.batch_size = batch_size
         self.epochs = epochs
         self.n_epochs_without_progress = n_epochs_without_progress
-        self.margin = margin
+        self.knn_distance_metric = knn_distance_metric
         self.ntrees = ntrees
         self.search_k = search_k
         self.precompute = precompute
-        self.model_def = model
+        self.model = model
         self.model_ = None
         self.encoder = None
         self.supervision_metric = supervision_metric
@@ -145,53 +163,50 @@ class Ivis(BaseEstimator):
             state['supervised_model_'] = None
         if 'callbacks' in state:
             state['callbacks'] = []
-        if not isinstance(state['model_def'], str):
-            state['model_def'] = None
+        if not isinstance(state['model'], str):
+            state['model'] = None
         if 'neighbour_matrix' in state:
             state['neighbour_matrix'] = None
+        if callable(state['distance']):
+            state['distance'] = state['distance'].__name__
         return state
 
     def _fit(self, X, Y=None, shuffle_mode=True):
 
-        if self.neighbour_matrix is not None:
-            datagen = generator_from_knn_matrix(X, Y,
-                                                neighbour_matrix=self.neighbour_matrix,
-                                                k=self.k,
-                                                batch_size=self.batch_size)
-        else:
+        if self.neighbour_matrix is None:
             if self.annoy_index_path is None:
                 self.annoy_index_path = 'annoy.index'
-                if self.verbose > 0:
-                    print('Building KNN index')
-                build_annoy_index(X, self.annoy_index_path,
-                                  ntrees=self.ntrees,
-                                  build_index_on_disk=self.build_index_on_disk,
-                                  verbose=self.verbose)
+                self.neighbour_matrix = AnnoyKnnMatrix.build(X, path=self.annoy_index_path,
+                                                             k=self.k, metric=self.knn_distance_metric,
+                                                             search_k=self.search_k,
+                                                             include_distances=False, ntrees=self.ntrees,
+                                                             build_index_on_disk=self.build_index_on_disk,
+                                                             precompute=self.precompute, verbose=self.verbose)
+            else:
+                self.neighbour_matrix = AnnoyKnnMatrix.load(self.annoy_index_path, X.shape,
+                                                            k=self.k, search_k=self.search_k,
+                                                            include_distances=False, precompute=self.precompute,
+                                                            verbose=self.verbose)
 
-            datagen = generator_from_index(X, Y,
-                                           index_path=self.annoy_index_path,
-                                           k=self.k,
-                                           batch_size=self.batch_size,
-                                           search_k=self.search_k,
-                                           precompute=self.precompute,
-                                           verbose=self.verbose)
+        datagen = generator_from_neighbour_matrix(X, Y,
+                                                  neighbour_matrix=self.neighbour_matrix,
+                                                  batch_size=self.batch_size)
 
         loss_monitor = 'loss'
         try:
-            triplet_loss_func = triplet_loss(distance=self.distance,
-                                             margin=self.margin)
+            triplet_loss_func = triplet_loss(distance=self.distance)
         except KeyError:
             raise ValueError('Loss function `{}` not implemented.'.format(self.distance))
 
         if self.model_ is None:
-            if isinstance(self.model_def, str):
+            if isinstance(self.model, str):
                 input_size = (X.shape[-1],)
                 self.model_, anchor_embedding, _, _ = \
-                    triplet_network(base_network(self.model_def, input_size),
+                    triplet_network(base_network(self.model, input_size),
                                     embedding_dims=self.embedding_dims)
             else:
                 self.model_, anchor_embedding, _, _ = \
-                    triplet_network(self.model_def,
+                    triplet_network(self.model,
                                     embedding_dims=self.embedding_dims)
 
             if Y is None:
@@ -283,8 +298,8 @@ class Ivis(BaseEstimator):
 
         Parameters
         ----------
-        X : n-dimensional array shape (n_samples, )
-            Data to be embedded.
+        X : np.array, ivis.data.sequence.IndexableDataset, tensorflow.keras.utils.HDF5Matrix
+            Data to be embedded. Needs to have a `.shape` attribute and a `__getitem__` method.
         Y : array, shape (n_samples)
             Optional array for supervised dimentionality reduction.
             If Y contains -1 labels, and 'sparse_categorical_crossentropy'
@@ -292,7 +307,8 @@ class Ivis(BaseEstimator):
 
         Returns
         -------
-        returns an instance of self
+        self: ivis.Ivis object
+            Returns estimator instance.
         """
 
         self._fit(X, Y, shuffle_mode)
@@ -303,8 +319,9 @@ class Ivis(BaseEstimator):
 
         Parameters
         ----------
-        X : array, shape (n_samples, n_features)
-            Data to be embedded.
+        X : np.array, ivis.data.sequence.IndexableDataset, tensorflow.keras.utils.HDF5Matrix
+            Data to train on and then embedded.
+            Needs to have a `.shape` attribute and a `__getitem__` method.
         Y : array, shape (n_samples)
             Optional array for supervised dimentionality reduction.
             If Y contains -1 labels, and 'sparse_categorical_crossentropy'
@@ -312,8 +329,8 @@ class Ivis(BaseEstimator):
 
         Returns
         -------
-        X_new : transformed array, shape (n_samples, embedding_dims)
-            Embedding of the new data in low-dimensional space.
+        X_new : array, shape (n_samples, embedding_dims)
+            Embedding of the data in low-dimensional space.
         """
 
         self.fit(X, Y, shuffle_mode)
@@ -325,16 +342,17 @@ class Ivis(BaseEstimator):
 
         Parameters
         ----------
-        X : array, shape (n_samples, n_features)
-            New data to be transformed.
+        X : np.array, ivis.data.sequence.IndexableDataset, tensorflow.keras.utils.HDF5Matrix
+            Data to be transformed. Needs to have a `.shape` attribute and a `__getitem__` method.
 
         Returns
         -------
         X_new : array, shape (n_samples, embedding_dims)
-            Embedding of the new data in low-dimensional space.
+            Embedding of the data in low-dimensional space.
         """
 
-        embedding = self.encoder.predict(X, verbose=self.verbose)
+        embedding = self.encoder.predict(KerasSequence(X, batch_size=self.batch_size),
+                                         verbose=self.verbose)
         return embedding
 
     def score_samples(self, X):
@@ -344,8 +362,9 @@ class Ivis(BaseEstimator):
 
         Parameters
         ----------
-        X : array, shape (n_samples, n_features)
+        X : np.array, ivis.data.sequence.IndexableDataset, tensorflow.keras.utils.HDF5Matrix
             Data to be passed through classification network.
+            Needs to have a `.shape` attribute and a `__getitem__` method.
 
         Returns
         -------
@@ -387,7 +406,7 @@ class Ivis(BaseEstimator):
 
         # save optimizer state in numpy array
         np.save(os.path.join(folder_path, 'optimizer_state.npy'),
-                self.model_.optimizer.get_weights())
+                np.array(self.model_.optimizer.get_weights(), dtype=object))
 
         json.dump(self.__getstate__(),
                   open(os.path.join(folder_path, 'ivis_params.json'), 'w'))
@@ -402,14 +421,17 @@ class Ivis(BaseEstimator):
 
         Returns
         -------
-        returns an ivis instance
+        self: ivis.Ivis object
+            Returns estimator instance.
         """
 
         ivis_config = json.load(open(os.path.join(folder_path,
                                                   'ivis_params.json'), 'r'))
+        if callable(self.distance):
+            ivis_config['distance'] = self.distance
         self.__dict__ = ivis_config
 
-        loss_function = triplet_loss(self.distance, self.margin)
+        loss_function = triplet_loss(self.distance)
 
         # ivis models trained before version 1.8.3 won't have separate optimizer file
         # maintain compatibility by falling back to old load behavior
